@@ -7,24 +7,27 @@ from __future__ import annotations
 
 import numbers
 import time
-from typing import TYPE_CHECKING, Any, Dict, List, Optional, Tuple, Type, Union
+from typing import TYPE_CHECKING, Any, Callable, Dict, List, Optional, Tuple, Type, Union
 
 import joblib
 import numpy as np
 from joblib import Memory
 from typing_extensions import Literal, TypedDict
 
-from tpcp._utils._general import _clone_parameter_dict, _get_nested_paras
+from tpcp._base import clone
+from tpcp._dataset import Dataset
+from tpcp._pipeline import Pipeline
+from tpcp._utils._general import _get_nested_paras
+
+if TYPE_CHECKING:
+    from tpcp._optimize import BaseOptimize
+    from tpcp.validate import Scorer
 
 _ERROR_SCORE_TYPE = Union[Literal["raise"], float]  # noqa: invalid-name
 _SCORE_TYPE = List[Union[Dict[str, float], float]]  # noqa: invalid-name
 _AGG_SCORE_TYPE = Union[Dict[str, float], float]  # noqa: invalid-name
 _SINGLE_SCORE_TYPE = Union[Dict[str, np.ndarray], np.ndarray]  # noqa: invalid-name
-
-if TYPE_CHECKING:
-    from tpcp import Dataset, SimplePipeline
-    from tpcp.optimize import BaseOptimize
-    from tpcp.validation import Scorer
+_SCORE_CALLABLE = Callable[[Pipeline, Dataset], Union[Dict[str, float], float]]  # noqa: invalid-name
 
 
 class ScoreResults(TypedDict, total=False):
@@ -53,7 +56,7 @@ class OptimizeScoreResults(TypedDict, total=False):
 
 
 def _score(
-    pipeline: SimplePipeline,
+    pipeline: Pipeline,
     dataset: Dataset,
     scorer: Scorer,
     parameters: Optional[Dict[str, Any]],
@@ -97,7 +100,7 @@ def _score(
             Calculated scores for each individual data point
         score_time : float
             Time required to score the dataset
-        data : List
+        data : list
             List of data point labels used
         parameters : dict or None
             The parameters that have been evaluated.
@@ -259,3 +262,11 @@ def _cached_optimize(
         )
 
     return optimizer
+
+
+def _clone_parameter_dict(param_dict: Optional[Dict]) -> Dict:
+    cloned_param_dict = {}
+    if param_dict is not None:
+        for k, v in param_dict.items():
+            cloned_param_dict[k] = clone(v, safe=False)
+    return cloned_param_dict
