@@ -6,42 +6,51 @@ from contextlib import nullcontext
 from functools import partial
 from itertools import product
 from tempfile import TemporaryDirectory
-from typing import Any, Callable, Dict, Iterator, List, Optional, Union
+from typing import Any, ContextManager, Dict, Iterator, List, Optional, TypeVar, Union
 
-import joblib
 import numpy as np
-from joblib import Memory, Parallel, delayed
+from joblib import Memory, delayed
 from numpy.ma import MaskedArray
 from scipy.stats import rankdata
 from sklearn.model_selection import BaseCrossValidator, ParameterGrid, check_cv
+from tqdm.auto import tqdm
 
-from tpcp import Dataset, OptimizablePipeline, SimplePipeline
+from tpcp import Dataset, OptimizablePipeline, Pipeline
+from tpcp._algorithm_utils import OPTIMIZE_METHOD_INDICATOR, _check_safe_optimize
+from tpcp._base import _get_annotated_fields_of_type
+from tpcp._optimize import BaseOptimize
+from tpcp._parameters import Parameter, _ParaTypes
 from tpcp._utils._general import (
     _aggregate_final_results,
     _normalize_score_results,
     _prefix_para_dict,
     _split_hyper_and_pure_parameters,
 )
-from tpcp._utils._multiprocess import init_progressbar
-from tpcp._utils._score import _optimize_and_score, _score
-from tpcp.base import BaseOptimize
+from tpcp._utils._multiprocess import TqdmParallel
+from tpcp._utils._score import _ERROR_SCORE_TYPE, _SCORE_CALLABLE, _optimize_and_score, _score
 from tpcp.exceptions import PotentialUserErrorWarning
-from tpcp.validation import Scorer
-from tpcp.validation._scorer import _ERROR_SCORE_TYPE, _validate_scorer
+from tpcp.validate import Scorer
+from tpcp.validate._scorer import _validate_scorer
+
+DummyOptimize_ = TypeVar("DummyOptimize_", bound="DummyOptimize")
+Optimize_ = TypeVar("Optimize_", bound="Optimize")
+GridSearch_ = TypeVar("GridSearch_", bound="GridSearch")
+GridSearchCv_ = TypeVar("GridSearchCv_", bound="GridSearchCV")
 
 
-class DummyOptimize(BaseOptimize):
+class DummyOptimize(BaseOptimize, _skip_validation=True):
     """Provide API compatibility for SimplePipelines in optimize wrappers.
 
-    This is a simple dummy Optimizer, that will **not** optimize anything, but just provide the correct api so that
-    pipelines that do not have the possibility to be optimizes can be passed to wrappers like `cross_validate`.
+    This is a simple dummy Optimizer that will **not** optimize anything, but just provide the correct API so that
+    pipelines that do not have the possibility to be optimized can be passed to wrappers like
+    :func:`tpcp.validate.cross_validate`.
 
 
     Parameters
     ----------
     pipeline
         The pipeline to wrap.
-        It will not optimized in any way, but simple copied to `self.optimized_pipeline_` if optimize is called.
+        It will not be optimized in any way, but simply copied to `self.optimized_pipeline_` if `optimize` is called.
 
     Other Parameters
     ----------------
@@ -53,18 +62,18 @@ class DummyOptimize(BaseOptimize):
     ----------
     optimized_pipeline_
         The optimized version of the pipeline.
-        In case of this class, this is just a unmodified clone of the input pipeline.
+        In case of this class, this is just an unmodified clone of the input pipeline.
 
     """
 
-    pipeline: SimplePipeline
+    pipeline: Parameter[Pipeline]
 
-    optimized_pipeline_: SimplePipeline
+    optimized_pipeline_: Pipeline
 
-    def __init__(self, pipeline: SimplePipeline):
+    def __init__(self, pipeline: Pipeline) -> None:  # noqa: super-init-not-called
         self.pipeline = pipeline
 
-    def optimize(self, dataset: Dataset, **optimize_params):
+    def optimize(self: DummyOptimize_, dataset: Dataset, **optimize_params: Any) -> DummyOptimize_:
         """Run the "dummy" optimization.
 
         Parameters
@@ -83,9 +92,9 @@ class DummyOptimize(BaseOptimize):
         self.dataset = dataset
         if hasattr(self.pipeline, "self_optimize"):
             warnings.warn(
-                "You are using `DummyOptimize` with a pipeline that implements `self_optimize` and hence indicates, "
+                "You are using `DummyOptimize` with a pipeline that implements `self_optimize` and, hence, indicates "
                 "that the pipeline can be optimized. "
-                "`DummmyOptimize` does never call this method and skips any optimization steps! "
+                "`DummyOptimize` does never call this method and skips any optimization steps! "
                 "Use `Optimize` if you actually want to optimize your pipeline.",
                 PotentialUserErrorWarning,
             )
@@ -98,16 +107,22 @@ class Optimize(BaseOptimize):
 
     This is a simple wrapper for pipelines that already implement a `self_optimize` method.
     This wrapper can be used to ensure that these algorithms can be optimized with the same interface as other
-    optimization methods and can hence be used in methods like cross-validation.
+    optimization methods and can hence be used in methods like :func:`tpcp.validate.cross_validate`.
 
     Optimize will never modify the original pipeline, but will store a copy of the optimized pipeline as
     `optimized_pipeline_`.
 
+    If `safe_optimize` is True, the wrapper applies the same runtime checks as provided by
+    :func:`~tpcp.make_optimize_safe`.
+
     Parameters
     ----------
     pipeline
-        The pipeline to optimize.
-        The pipeline must implement `self_optimize` to optimize its own input parameters.
+        The pipeline to optimize. The pipeline must implement `self_optimize` to optimize its own input parameters.
+    safe_optimize
+        If True, we add additional checks to make sure the `self_optimize` method of the pipeline is correctly
+        implemented.
+        See :func:`~tpcp.make_optimize_safe` for more info.
 
     Other Parameters
     ----------------
@@ -122,14 +137,18 @@ class Optimize(BaseOptimize):
 
     """
 
-    pipeline: OptimizablePipeline
+    pipeline: Parameter[OptimizablePipeline]
+    safe_optimize: bool
 
     optimized_pipeline_: OptimizablePipeline
 
-    def __init__(self, pipeline: OptimizablePipeline):
+    def __init__(  # noqa: super-init-not-called
+        self, pipeline: OptimizablePipeline, *, safe_optimize: bool = True
+    ) -> None:
         self.pipeline = pipeline
+        self.safe_optimize = safe_optimize
 
-    def optimize(self, dataset: Dataset, **optimize_params):
+    def optimize(self: Optimize_, dataset: Dataset, **optimize_params: Any) -> Optimize_:
         """Run the self-optimization defined by the pipeline.
 
         The optimized version of the pipeline is stored as `self.optimized_pipeline_`.
@@ -155,45 +174,38 @@ class Optimize(BaseOptimize):
             raise ValueError(
                 "To use `Optimize` with a pipeline, the pipeline needs to implement a `self_optimize` method."
             )
-        # record the hash of the pipeline to make an educated guess if the optimization works
-        pipeline = self.pipeline.clone()
-        before_hash = joblib.hash(pipeline)
-        optimized_pipeline = pipeline.self_optimize(dataset, **optimize_params)
-        if not isinstance(optimized_pipeline, pipeline.__class__):
-            raise ValueError(
-                "Calling `self_optimize` did not return an instance of the pipeline itself! "
-                "Normally this method should return `self`."
-            )
-        # We clone the optimized pipeline again, to make sure that only changes to the input parameters are kept.
-        optimized_pipeline = optimized_pipeline.clone()
-        after_hash = joblib.hash(optimized_pipeline)
-        if before_hash == after_hash:
-            # If the hash didn't change the object didn't change.
-            # Something might have gone wrong.
-            warnings.warn(
-                "Optimizing the pipeline doesn't seem to have changed the parameters of the pipeline. "
-                "This could indicate an implementation error of the `self_optimize` method.",
-                PotentialUserErrorWarning,
-            )
-        self.optimized_pipeline_ = optimized_pipeline
+        # We clone just to make sure runs are independent
+        pipeline: OptimizablePipeline = self.pipeline.clone()
+        if self.safe_optimize is True:
+            # We check here, if the pipeline already has the safe decorator and if yes just call it.
+            if getattr(pipeline.self_optimize, OPTIMIZE_METHOD_INDICATOR, False) is True:
+                optimized_pipeline = pipeline.self_optimize(dataset, **optimize_params)
+            else:
+                optimized_pipeline = _check_safe_optimize(pipeline, pipeline.self_optimize, dataset, **optimize_params)
+        else:
+            optimized_pipeline = pipeline.self_optimize(dataset, **optimize_params)
+        # We clone again, just to be sure
+        self.optimized_pipeline_ = optimized_pipeline.clone()
         return self
 
 
 class GridSearch(BaseOptimize):
-    """Perform a GridSearch over various parameters.
+    """Perform a grid search over various parameters.
 
-    This scores the pipeline for every combination of data-points in the provided dataset and parameter combinations
+    This scores the pipeline for every combination of data points in the provided dataset and parameter combinations
     in the `parameter_grid`.
-    The scores over the entire dataset are then aggregated for each para combination.
-    By default this aggregation is a simple average.
+    The scores over the entire dataset are then aggregated for each parameter combination.
+    By default, this aggregation is a simple average.
 
-    Note, that this is different to how GridSearch works in many other cases:
-    Usually, the performance parameter would be calculated on all data-points at once.
-    Here, each data-point represents an entire participant or gait-recording (depending on the dataset).
-    Therefore, the pipeline and the scoring method are expected to provide a result/score per data-point in the dataset.
-    Note, that it is still open to your interpretation, what you consider a datapoint in the context of your analysis.
-    The run method of the pipeline can still process multiple e.g. gaittests in a loop and generate a single output,
-    if you consider a single participant one datapoint.
+    .. note::
+        This is different to how grid search works in many other cases:
+        Usually, the performance parameter would be calculated on all data points at once.
+        Here, each data point represents an entire participant or recording (depending on the dataset).
+        Therefore, the pipeline and the scoring method are expected to provide a result/score per data point
+        in the dataset.
+        Note that it is still open to your interpretation what you consider a "data point" in the context of your
+        analysis. The `run` method of the pipeline can still process multiple data points, e.g., gait tests, in a loop
+        and generate a single output if you consider a single participant *one data point*.
 
     Parameters
     ----------
@@ -206,17 +218,17 @@ class GridSearch(BaseOptimize):
         This function should return either a single score or a dictionary of scores.
         If scoring is `None` the default `score` method of the pipeline is used instead.
 
-        .. note:: If scoring returns a dictionary, `return_optimized` must be set to the name of the score that
-                  should be used for ranking.
+        Note that if scoring returns a dictionary, `return_optimized` must be set to the name of the score that
+        should be used for ranking.
     n_jobs
         The number of processes that should be used to parallelize the search.
-        None means 1, -1 means as many as logical processing cores.
+        `None` means 1 while -1 means as many as logical processing cores.
     pre_dispatch
         The number of jobs that should be pre dispatched.
         For an explanation see the documentation of :class:`~sklearn.model_selection.GridSearchCV`
     return_optimized
         If True, a pipeline object with the overall best params is created and stored as `optimized_pipeline_`.
-        If `scoring` returns returns a dictionary of score values, this must be a str corresponding to the name of the
+        If `scoring` returns a dictionary of score values, this must be a `str` corresponding to the name of the
         score that should be used to rank the results.
         If False, the respective result attributes will not be populated.
         If multiple parameter combinations have the same score, the one tested first will be used.
@@ -226,7 +238,7 @@ class GridSearch(BaseOptimize):
         If set to ‘raise’, the error is raised.
         If a numeric value is given, a Warning is raised.
     progress_bar
-        True/False to enable/disable a tqdm progressbar.
+        True/False to enable/disable a tqdm progress bar.
 
     Other Parameters
     ----------------
@@ -237,7 +249,7 @@ class GridSearch(BaseOptimize):
     ----------
     gs_results_
         A dictionary summarizing all results of the gridsearch.
-        The format of this dictionary is designed to be directly passed into the `pd.DataFrame` constructor.
+        The format of this dictionary is designed to be directly passed into the :class:`~pd.DataFrame` constructor.
         Each column then represents the result for one set of parameters
 
         The dictionary contains the following entries:
@@ -248,16 +260,16 @@ class GridSearch(BaseOptimize):
             A dictionary representing all parameters
         score / {scorer-name}
             The aggregated value of a score over all data-points.
-            If a single score is used for scoring, than the generic name "score" is used.
-            Otherwise multiple columns with the name of the respective scorer exist
+            If a single score is used for scoring, then the generic name "score" is used.
+            Otherwise, multiple columns with the name of the respective scorer exist
         rank_score / rank_{scorer-name}
             A sorting for each score from the highest to the lowest value
         single_score / single_{scorer-name}
-            The individual scores per datapoint for each parameter combination.
+            The individual scores per data point for each parameter combination.
             This is a list of values with the `len(dataset)`.
         data_labels
             A list of data labels in the order the single score values are provided.
-            These can be used to associate the `single_score` values with a certain data-point.
+            These can be used to associate the `single_score` values with a certain data point.
     optimized_pipeline_
         An instance of the input pipeline with the best parameter set.
         This is only available if `return_optimized` is not False.
@@ -277,7 +289,7 @@ class GridSearch(BaseOptimize):
     """
 
     parameter_grid: ParameterGrid
-    scoring: Optional[Union[Callable, Scorer]]
+    scoring: Optional[Union[_SCORE_CALLABLE, Scorer]]
     n_jobs: Optional[int]
     return_optimized: Union[bool, str]
     pre_dispatch: Union[int, str]
@@ -285,23 +297,23 @@ class GridSearch(BaseOptimize):
     progress_bar: bool
 
     gs_results_: Dict[str, Any]
-    best_params_: Dict
+    best_params_: Dict[str, Any]
     best_index_: int
     best_score_: float
     multimetric_: bool
 
-    def __init__(
+    def __init__(  # noqa: super-init-not-called
         self,
-        pipeline: SimplePipeline,
+        pipeline: Pipeline,
         parameter_grid: ParameterGrid,
         *,
-        scoring: Optional[Union[Callable, Scorer]] = None,
+        scoring: Optional[Union[_SCORE_CALLABLE, Scorer]] = None,
         n_jobs: Optional[int] = None,
-        pre_dispatch: Union[int, str] = "n_jobs",
         return_optimized: Union[bool, str] = True,
+        pre_dispatch: Union[int, str] = "n_jobs",
         error_score: _ERROR_SCORE_TYPE = np.nan,
         progress_bar: bool = True,
-    ):
+    ) -> None:
         self.pipeline = pipeline
         self.parameter_grid = parameter_grid
         self.scoring = scoring
@@ -311,8 +323,8 @@ class GridSearch(BaseOptimize):
         self.error_score = error_score
         self.progress_bar = progress_bar
 
-    def optimize(self, dataset: Dataset, **_):
-        """Run the GridSearch over the dataset and find the best parameter combination.
+    def optimize(self: GridSearch_, dataset: Dataset, **_: Any) -> GridSearch_:
+        """Run the grid search over the dataset and find the best parameter combination.
 
         Parameters
         ----------
@@ -323,41 +335,37 @@ class GridSearch(BaseOptimize):
         self.dataset = dataset
         scoring = _validate_scorer(self.scoring, self.pipeline)
 
-        # We use a similar structure as sklearns GridSearchCV here, but instead of calling something equivalent to
+        # We use a similar structure as sklearn's GridSearchCV here, but instead of calling something equivalent to
         # `fit_score`, we call `score`, which just applies and scores the pipeline on the entirety of our dataset as
         # we do not need a "train" step.
-        # Our main loop just loops over all parameter combis and the `_score` function then applies the para combi to
-        # the pipeline and scores the resulting pipeline on the dataset, by passing the entire dataset and the
-        # pipeline to the scorer.
-        # Looping over the individual datapoints in the dataset and aggregating the scores is handled by the scorer
+        # Our main loop just loops over all parameter combinations and the `_score` function then applies the parameter
+        # combination to the pipeline and scores the resulting pipeline on the dataset, by passing the entire dataset
+        # and the pipeline to the scorer.
+        # Looping over the individual data points in the dataset and aggregating the scores is handled by the scorer
         # itself.
         # If not explicitly changed the scorer is an instance of `Scorer` that wraps the actual `scoring`
         # function provided by the user.
-        with init_progressbar(
-            self.progress_bar,
-            n_jobs=self.n_jobs,
-            iterable=self.parameter_grid,
-            desc="Para Combos",
-            total=len(self.parameter_grid),
-        ) as wrapped_iterable:
-            parallel = Parallel(n_jobs=self.n_jobs, pre_dispatch=self.pre_dispatch)
-            with parallel:
-                # Evaluate each parameter combination
-                results = parallel(
-                    delayed(_score)(
-                        self.pipeline.clone(),
-                        dataset,
-                        scoring,
-                        paras,
-                        return_parameters=True,
-                        return_data_labels=True,
-                        return_times=True,
-                        error_score=self.error_score,
-                    )
-                    for paras in wrapped_iterable
+        pbar: Optional[tqdm] = None
+        if self.progress_bar:
+            pbar = tqdm(total=len(self.parameter_grid), desc="Parameter Combinations")
+        parallel = TqdmParallel(n_jobs=self.n_jobs, pre_dispatch=self.pre_dispatch, pbar=pbar)
+        with parallel:
+            # Evaluate each parameter combination
+            results = parallel(
+                delayed(_score)(
+                    self.pipeline.clone(),
+                    dataset,
+                    scoring,
+                    paras,
+                    return_parameters=True,
+                    return_data_labels=True,
+                    return_times=True,
+                    error_score=self.error_score,
                 )
+                for paras in self.parameter_grid
+            )
         # We check here if all results are dicts. We only check the dtype of the first value, as the scorer should
-        # have handled issues with non uniform cases already.
+        # have handled issues with non-uniform cases already.
         first_test_score = results[0]["scores"]
         self.multimetric_ = isinstance(first_test_score, dict)
         _validate_return_optimized(self.return_optimized, self.multimetric_, first_test_score)
@@ -384,7 +392,7 @@ class GridSearch(BaseOptimize):
     def _format_results(self, candidate_params, out):  # noqa: no-self-use
         """Format the final result dict.
 
-        This function is adapted based on sklearns `BaseSearchCV`
+        This function is adapted based on sklearn's `BaseSearchCV`
         """
         n_candidates = len(candidate_params)
         out = _aggregate_final_results(out)
@@ -429,12 +437,12 @@ class GridSearch(BaseOptimize):
 
 
 class GridSearchCV(BaseOptimize):
-    """Exhaustive (Hyper)Parameter search using a cross validation based score to optimize pipeline parameters.
+    """Exhaustive (hyper)parameter search using a cross validation based score to optimize pipeline parameters.
 
     This class follows as much as possible the interface of :func:`~sklearn.model_selection.GridSearchCV`.
-    If the tpcp documentation is missing some information, the respective documentation of sklearn might be helpful.
+    If the `tpcp` documentation is missing some information, the respective documentation of `sklearn` might be helpful.
 
-    Compared to the sklearn implementation this method uses a couple of tpcp specific optimizations and
+    Compared to the `sklearn` implementation this method uses a couple of `tpcp`-specific optimizations and
     quality-of-life improvements.
 
     Parameters
@@ -451,41 +459,49 @@ class GridSearchCV(BaseOptimize):
         .. note:: If scoring returns a dictionary, `return_optimized` must be set to the name of the score that
                   should be used for ranking.
     return_optimized
-        If True, a pipeline object with the overall best params is created  and reoptimized using all provided data
+        If True, a pipeline object with the overall best parameters is created and re-optimized using all provided data
         as input.
         The optimized pipeline object is stored as `optimized_pipeline_`.
-        If `scoring` returns returns a dictionary of score values, this must be a str corresponding to the name of the
+        If `scoring` returns a dictionary of score values, this must be a str corresponding to the name of the
         score that should be used to rank the results.
         If False, the respective result attributes will not be populated.
         If multiple parameter combinations have the same mean score over all CV folds, the one tested first will be
         used.
         Otherwise, higher mean values are always considered better.
     cv
-        An integer specifying the number of folds in a K-Fold cross-validation or a valid cross validation helper.
+        An integer specifying the number of folds in a K-Fold cross validation or a valid cross validation helper.
         The default (`None`) will result in a 5-fold cross validation.
-        For further inputs check the sklearn documentation.
-    pure_parameter_names
-        .. warning: Do not use this option unless you fully understand it!
+        For further inputs check the `sklearn`
+        `documentation <https://scikit-learn.org/stable/modules/generated/sklearn.model_selection.GridSearchCV.html>`_.
+    pure_parameters
+        .. warning::
+            Do not use this option unless you fully understand it!
 
-        A list of parameter names (named in the `parameter_grid`) that do not effect training aka are not
-        Hyperparameters.
+        A list of parameter names (named in the `parameter_grid`) that do not affect training aka are not
+        hyperparameters.
         This information can be used for massive performance improvements, as the training does not need to be
-        repeated, if one of these parameters changes.
-        However, setting it incorrectly can lead to hard to detect errors in the final results.
+        repeated if one of these parameters changes.
+        However, setting it incorrectly can lead detect errors that are very hard to detect in the final results.
 
-        For more information on this approach see the :ref:`evaluation guide <algorithm_evaluation>`.
+        Instead of passing a list of names, you can also just set the value to `True`.
+        In this case all parameters of the provided pipeline that are marked as :func:`~tpcp.pure_parameter` are used.
+        Note that pure parameters of nested objects are not considered, but only top-level attributes.
+        If you need to mark nested parameters as pure, use the first method and pass the names (with `__`) as part of
+        the list of names.
+
+        For more information on this approach see the :func:`evaluation guide <algorithm_evaluation>`.
     return_train_score
         If True the performance on the train score is returned in addition to the test score performance.
         Note, that this increases the runtime.
         If `True`, the fields `train_score`, and `train_score_single` are available in the results.
     verbose
         Control the verbosity of information printed during the optimization (larger number -> higher verbosity).
-        At the moment this will only effect the caching done, when `pure_parameter_names` are provided.
+        At the moment this will only affect the caching done, when `pure_parameter_names` are provided.
     n_jobs
         The number of parallel jobs.
         The default (`None`) means 1 job at the time, hence, no parallel computing.
         -1 means as many as logical processing cores.
-        One job is created per cv + para combi combintion.
+        One job is created per cv + para combi combination.
     pre_dispatch
         The number of jobs that should be pre dispatched.
         For an explanation see the documentation of :class:`~sklearn.model_selection.GridSearchCV`
@@ -495,6 +511,10 @@ class GridSearchCV(BaseOptimize):
         If a numeric value is given, a Warning is raised.
     progress_bar
         True/False to enable/disable a tqdm progressbar.
+    safe_optimize
+        If True, we add additional checks to make sure the `self_optimize` method of the pipeline is correctly
+        implemented.
+        See :func:`~tpcp.make_optimize_safe` for more info.
 
     Other Parameters
     ----------------
@@ -525,9 +545,9 @@ class GridSearchCV(BaseOptimize):
         split{n}_test_score / split{n}_test_{scorer_name}
             The performance on the test set in fold n.
         split{n}_test_single_score / split{n}_test_single_{scorer_name}
-            The performance in fold n on every single datapoint in the test set.
+            The performance in fold n on every single data point in the test set.
         split{n}_test_data_labels
-            The ids of the datapoints used in the test set of fold n.
+            The ids of the data points used in the test set of fold n.
         mean_train_score / mean_train_{scorer_name}
             The average train score over all folds.
         std_train_score / std_train_{scorer_name}
@@ -539,7 +559,7 @@ class GridSearchCV(BaseOptimize):
         split{n}_train_single_score / split{n}_train_single_{scorer_name}
             The performance in fold n on every single datapoint in the train set.
         split{n}_train_data_labels
-            The ids of the datapoints used in the train set of fold n.
+            The ids of the data points used in the train set of fold n.
         mean_{optimize/score}_time
             Average time over all folds spent for optimization and scoring, respectively.
         std_{optimize/score}_time
@@ -568,54 +588,58 @@ class GridSearchCV(BaseOptimize):
 
     pipeline: OptimizablePipeline
     parameter_grid: ParameterGrid
-    scoring: Optional[Union[Callable, Scorer]]
+    scoring: Optional[Union[_SCORE_CALLABLE, Scorer]]
     return_optimized: Union[bool, str]
     cv: Optional[Union[int, BaseCrossValidator, Iterator]]
-    pure_parameter_names: Optional[List[str]]
+    pure_parameters: Union[bool, List[str]]
     return_train_score: bool
     verbose: int
     n_jobs: Optional[int]
     pre_dispatch: Union[int, str]
     error_score: _ERROR_SCORE_TYPE
     progress_bar: bool
+    safe_optimize: bool
 
     cv_results_: Dict[str, Any]
-    best_params_: Dict
+    best_params_: Dict[str, Any]
     best_index_: int
     best_score_: float
     multimetric_: bool
     final_optimize_time_: float
 
-    def __init__(
+    def __init__(  # noqa: super-init-not-called
         self,
         pipeline: OptimizablePipeline,
         parameter_grid: ParameterGrid,
         *,
-        scoring: Optional[Union[Callable, Scorer]] = None,
+        scoring: Optional[Union[_SCORE_CALLABLE, Scorer]] = None,
         return_optimized: Union[bool, str] = True,
         cv: Optional[Union[int, BaseCrossValidator, Iterator]] = None,
-        pure_parameter_names: Optional[List[str]] = None,
+        pure_parameters: Union[bool, List[str]] = False,
         return_train_score: bool = False,
         verbose: int = 0,
         n_jobs: Optional[int] = None,
         pre_dispatch: Union[int, str] = "n_jobs",
         error_score: _ERROR_SCORE_TYPE = np.nan,
         progress_bar: bool = True,
-    ):
+        safe_optimize: bool = True,
+    ) -> None:
         self.pipeline = pipeline
         self.parameter_grid = parameter_grid
         self.scoring = scoring
         self.return_optimized = return_optimized
         self.cv = cv
-        self.pure_parameter_names = pure_parameter_names
+        self.pure_parameters = pure_parameters
         self.return_train_score = return_train_score
         self.verbose = verbose
         self.n_jobs = n_jobs
         self.pre_dispatch = pre_dispatch
         self.error_score = error_score
         self.progress_bar = progress_bar
+        self.safe_optimize = safe_optimize
 
-    def optimize(self, dataset: Dataset, *, groups=None, **optimize_params):  # noqa: arguments-differ
+    def optimize(self: GridSearchCv_, dataset: Dataset, *, groups=None, **optimize_params) -> GridSearchCv_:  # noqa:
+        # arguments-differ
         self.dataset = dataset
         scoring = _validate_scorer(self.scoring, self.pipeline)
 
@@ -624,56 +648,65 @@ class GridSearchCV(BaseOptimize):
 
         # We need to wrap our pipeline for a consistent interface.
         # In the future we might be able to allow objects with optimizer Interface as input directly.
-        optimizer = Optimize(self.pipeline)
+        optimizer = Optimize(self.pipeline, safe_optimize=self.safe_optimize)
 
-        # For each para combi, we separate the pure parameters (parameters that do not effect the optimization) and
+        # For each para combi, we separate the pure parameters (parameters that do not affect the optimization) and
         # the hyperparameters.
         # This allows for massive caching optimizations in the `_optimize_and_score`.
+        pure_parameters: Optional[List[str]]
+        if self.pure_parameters is False:
+            pure_parameters = None
+        elif self.pure_parameters is True:
+            pure_parameters = _get_annotated_fields_of_type(self.pipeline, _ParaTypes.PURE)
+        elif isinstance(self.pure_parameters, list):
+            pure_parameters = self.pure_parameters
+        else:
+            raise ValueError(
+                "`self.pure_parameters` must either be a List of field names (nested are allowed) or " "True/False."
+            )
+
         parameters = list(self.parameter_grid)
-        split_parameters = _split_hyper_and_pure_parameters(parameters, self.pure_parameter_names)
+        split_parameters = _split_hyper_and_pure_parameters(parameters, pure_parameters)
         parameter_prefix = "pipeline__"
+        combinations = list(product(enumerate(split_parameters), enumerate(cv.split(dataset, groups=groups))))
+
+        pbar: Optional[tqdm] = None
+        if self.progress_bar:
+            pbar = tqdm(total=len(combinations), desc="Split-Para Combos")
 
         # To enable the pure parameter performance improvement, we need to create a joblib cache in a temp dir that
         # is deleted after the run.
         # We only allow a temporary cache here, because the method that is cached internally is generic and the cache
         # might not be correctly invalidated, if GridSearchCv is called with a different pipeline or when the
         # pipeline itself is modified.
-        tmp_dir_context = nullcontext()
-        if self.pure_parameter_names:
+        tmp_dir_context: Union[ContextManager[None], TemporaryDirectory] = nullcontext()
+        if pure_parameters:
             tmp_dir_context = TemporaryDirectory("joblib_tpcp_cache")
         with tmp_dir_context as cachedir:
             tmp_cache = Memory(cachedir, verbose=self.verbose) if cachedir else None
-            combinations = list(product(enumerate(split_parameters), enumerate(cv.split(dataset, groups=groups))))
-            with init_progressbar(
-                self.progress_bar,
-                n_jobs=self.n_jobs,
-                iterable=combinations,
-                desc="Split-Para Combos",
-                total=len(combinations),
-            ) as wrapped_iterable:
-                parallel = Parallel(n_jobs=self.n_jobs, pre_dispatch=self.pre_dispatch)
-                # We use a similar structure to sklearns GridSearchCv here (see GridSearch for more info).
-                with parallel:
-                    # Evaluate each parameter combination
-                    out = parallel(
-                        delayed(_optimize_and_score)(
-                            optimizer.clone(),
-                            dataset,
-                            scoring,
-                            train,
-                            test,
-                            optimize_params=optimize_params,
-                            hyperparameters=_prefix_para_dict(hyper_paras, parameter_prefix),
-                            pure_parameters=_prefix_para_dict(pure_paras, parameter_prefix),
-                            return_train_score=self.return_train_score,
-                            return_parameters=False,
-                            return_data_labels=True,
-                            return_times=True,
-                            error_score=self.error_score,
-                            memory=tmp_cache,
-                        )
-                        for (cand_idx, (hyper_paras, pure_paras)), (split_idx, (train, test)) in wrapped_iterable
+            parallel = TqdmParallel(n_jobs=self.n_jobs, pre_dispatch=self.pre_dispatch, pbar=pbar)
+            # We use a similar structure to sklearn's GridSearchCV here (see GridSearch for more info).
+            with parallel:
+                # Evaluate each parameter combination
+                out = parallel(
+                    delayed(_optimize_and_score)(
+                        optimizer.clone(),
+                        dataset,
+                        scoring,
+                        train,
+                        test,
+                        optimize_params=optimize_params,
+                        hyperparameters=_prefix_para_dict(hyper_paras, parameter_prefix),
+                        pure_parameters=_prefix_para_dict(pure_paras, parameter_prefix),
+                        return_train_score=self.return_train_score,
+                        return_parameters=False,
+                        return_data_labels=True,
+                        return_times=True,
+                        error_score=self.error_score,
+                        memory=tmp_cache,
                     )
+                    for (cand_idx, (hyper_paras, pure_paras)), (split_idx, (train, test)) in combinations
+                )
         results = self._format_results(parameters, n_splits, out)
         self.cv_results_ = results
 
@@ -699,7 +732,7 @@ class GridSearchCV(BaseOptimize):
     def _format_results(self, candidate_params, n_splits, out, more_results=None):  # noqa: MC0001
         """Format the final result dict.
 
-        This function is adapted based on sklearns `BaseSearchCV`.
+        This function is adapted based on sklearn's `BaseSearchCV`.
         """
         n_candidates = len(candidate_params)
         out = _aggregate_final_results(out)
@@ -756,7 +789,7 @@ class GridSearchCV(BaseOptimize):
         # Use one MaskedArray and mask all the places where the param is not
         # applicable for that candidate. Use defaultdict as each candidate may
         # not contain all the params
-        param_results = defaultdict(
+        param_results: Dict = defaultdict(
             partial(
                 MaskedArray,
                 np.empty(
@@ -769,7 +802,7 @@ class GridSearchCV(BaseOptimize):
         for cand_idx, params in enumerate(candidate_params):
             for name, value in params.items():
                 # An all masked empty array gets created for the key
-                # `"param_{name}` at the first occurrence of `name`.
+                # `"param_{name}"` at the first occurrence of `name`.
                 # Setting the value at an index also unmasks that index
                 param_results[f"param_{name}"][cand_idx] = value
 
@@ -794,7 +827,7 @@ class GridSearchCV(BaseOptimize):
         return results
 
 
-def _validate_return_optimized(return_optimized, multi_metric, results):
+def _validate_return_optimized(return_optimized, multi_metric, results) -> None:
     """Check if `return_optimize` fits to the multimetric output of the scorer."""
     if multi_metric is True:
         # In a multimetric case, return_optimized must either be False or a string
